/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_ext.h

   Falcon module for Apache 2
   Definition of apache-specific module.
   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-07-28 18:55:17

   -------------------------------------------------------------------
   (C) Copyright 2008: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

#include "mod_falcon_config.h"
#include "apache_ext.h"
#include "apache_vm.h"
#include "apache_output.h"
#include "apache_stream.h"
#include "version.h"
#include <stdlib.h>
#include <falcon/engine.h>
#include <falcon/fstream.h>
#include <falcon/sys.h>
#include <util_filter.h>

#include <apr_strings.h>
#include <http_log.h>

#define BOUNDARY_SIZE   128

static void s_random_filename( char *buffer, int len )
{
   //TODO: Use APR Random routines.
   static const char* alphabeth= "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
   
   for( int i = 0; i < len; i ++ ){
      buffer[i] = alphabeth[random() % 63];
   }
   buffer[len] = '\0';
}

static void internal_htmlEscape_stream( const Falcon::String &str, Falcon::Stream *out )
{
   for ( Falcon::uint32 i = 0; i < str.length(); i++ )
   {
      Falcon::uint32 chr = str[i];
      switch ( chr )
      {
         case '<': out->writeString( "&lt;" ); break;
         case '>': out->writeString( "&gt;" ); break;
         case '"': out->writeString( "&quot;" ); break;
         case '&': out->writeString( "&amp;" ); break;
         default:
            out->put( chr );
      }
   }
}


static void s_dict_as_uri_string( Falcon::Sequence* items, Falcon::String& target )
{
   Falcon::Iterator iter( items );
   while( iter.hasCurrent() )
   {
      const Falcon::Item &key = iter.getCurrentKey();
      const Falcon::Item &value = iter.getCurrent();

      if( key.isString() && value.isString() )
      {
         if ( target.size() != 0 )
            target += "&";
         Falcon::String temp;
         Falcon::URI::URLEncode( *key.asString(), temp );
         target += temp;
         target += "=";
         Falcon::URI::URLEncode( *value.asString(), temp );
         target += temp;
      }

      iter.next();
   }
}

static void s_dict_as_input_fields( Falcon::Sequence* items, Falcon::Stream *out )
{
   Falcon::Iterator iter( items );
   while( iter.hasCurrent() )
   {
      const Falcon::Item &key = iter.getCurrentKey();
      const Falcon::Item &value = iter.getCurrent();

      if( key.isString() && value.isString() )
      {
         out->writeString( "<input type=\"hidden\" name=\"" );
         internal_htmlEscape_stream( *key.asString(), out );
         out->writeString( "\" value=\"" );
         internal_htmlEscape_stream( *value.asString(), out );
         out->writeString( "\"/>\n" );
      }

      iter.next();
   }
}


//=====================================================================
// Request
//
struct table_callback {
   ApacheVM *avm;
   Falcon::CoreDict *headers;
   Falcon::CoreDict *cookies;
   // multipart management
   bool bIsMultiPart;
   int contentLength;
   char boundary[BOUNDARY_SIZE];
   int boundaryLen;
};

static int table_to_dict(void *rec, const char *key, const char *value)
{
   struct table_callback *tbc = (struct table_callback *) rec;
   if ( apr_strnatcasecmp( key, "Cookie" ) == 0 )
   {
      // break the cookie.
      // we have a line.
      char *parse_value = apr_pstrdup( tbc->avm->request()->pool, value);
      char *last;
      char *token = apr_strtok(parse_value,";",&last);
      while( token != NULL )
      {
         // break the "="
         char *keylast;
         char *keytok = apr_strtok(token,"=",&keylast);
         if ( keytok != NULL )
         {
            char *valtok = apr_strtok(NULL, "=",&keylast);
            if ( valtok != NULL )
            {
               // Ok, only in this case our cookie is regular.
               // Now perform stringization of key and value.
               Falcon::CoreString *skey = new Falcon::CoreString;
               while( *keytok == ' ' ) keytok++;

               if ( Falcon::URI::URLDecode( keytok, *skey ) )
               {
                  Falcon::CoreString *svalue = new Falcon::CoreString;
                  while( *valtok == ' ' ) valtok++;
                  if ( Falcon::URI::URLDecode( valtok, *svalue ) )
                  {
                     tbc->cookies->put(skey, svalue);
                  }
               }
            }
         }
         token = apr_strtok(NULL,";",&last);
      }
   }
   else
   {
      // anyhow, add the field.
      tbc->headers->put( new Falcon::CoreString( (const char*) key, (Falcon::int32) strlen( key ) ),
         new Falcon::CoreString( (const char*)value, (Falcon::int32) strlen( value ) ));

      // then check it
      if ( apr_strnatcasecmp( key, "Content-Type" ) == 0 )
      {
         // may be a multipart. -- if it is, boundary is mandatory.
         if( strncmp( value, "multipart/form-data;", 20 ) == 0 )
         {
            // scan for the boundary key
            const char *keytok = value + 20;
            while( *keytok == ' ' ) keytok++;

            if( strncmp( keytok, "boundary", 8 ) == 0 )
            {
               keytok += 8;
               const char *base = keytok;
               while( *keytok == ' ' || *keytok == '=' ) keytok++;

               // this ensures we have matched a "boundary =" and not i.e. a "boundary-nice"
               if ( keytok != base )
               {
                  // the real boundary has an extra "--" in front (as specified by mime)
                  // Also, it is preceded and followed by \r\n, but we don't precede it now,
                  // as we need a separate \r\n marker for line-based parts in multipart docs.
                  apr_cpystrn( tbc->boundary+2, keytok, BOUNDARY_SIZE-2 );
                  tbc->boundary[0] = '-';
                  tbc->boundary[1] = '-';
                  tbc->boundaryLen = strlen( tbc->boundary );
                  tbc->bIsMultiPart = true;
               }
            }
            // else ignore this field
         }
      }
      else if ( apr_strnatcasecmp( key, "Content-Length" ) == 0 )
      {
         tbc->contentLength = atoi( value );
      }
   }

   return 1;
}

// now parse the post request.
static int s_read_post( Falcon::String &qs, apr_bucket_brigade *brigade, request_rec *r  )
{
   apr_size_t len, tlen=0;
   char buf[1025];
   len = 1024;

   /*
    * This loop is needed because ap_get_brigade() can return us partial data
    * which would cause premature termination of request read. Therefor we
    * need to make sure that if data is available we fill the buffer completely.
    */

   while (ap_get_brigade(r->input_filters, brigade, AP_MODE_READBYTES, APR_BLOCK_READ, len) == APR_SUCCESS)
   {
      apr_brigade_flatten(brigade, buf, &len);
      apr_brigade_cleanup(brigade);
      if ( !len) {
         break;
      }
      tlen += len;
      buf[len] = 0;
      qs.append( buf );
      len = 1024;
   }

   return tlen;
}

void s_unescape_quotes( Falcon::String *str )
{
   str->trim();
   Falcon::uint32 len = str->length();
   if ( len > 1 )
   {
      if( str->getCharAt(0) == '"' && str->getCharAt(len-1) == '"' )
      {
         str->remove(0,1);
         str->remove( len-2, 1 );
         len -= 2;
         bool bEscape;
         Falcon::uint32 i = 0;
         while( i + 1 < len )
         {
            if ( str->getCharAt(i) == '\\' &&  str->getCharAt(i+1) == '"' )
            {
               str->remove( i, 1 );
            }
            i++;
         }
      }
   }
}

Falcon::CoreString* s_get_header_field( const char* field, Falcon::VMachine* vm )
{
   Falcon::CoreString* ret = new Falcon::CoreString;
   ret->fromUTF8( field );
   s_unescape_quotes( ret );
   return ret;
}

// Reads a single field in the request
bool s_read_uploaded_file( Falcon::VMachine *vm, struct table_callback &tbc,
                     apr_bucket_brigade *brigade, request_rec *r, Falcon::CoreObject* upld )
{
   apr_size_t len, tlen=0;
   char buf[2048];
   char fname[MAX_UPLOAD_DIR + 20];
   len = 2046;

   apr_file_t *fp_temp = 0;
   
   char boundary[ BOUNDARY_SIZE + 2 ];
   memcpy( boundary + 2, tbc.boundary, tbc.boundaryLen+1 );
   boundary[0] = '\r';
   boundary[1] = '\n';
   int bl = tbc.boundaryLen+2;

   // should we write to a file or store in memory?
   Falcon::MemBuf *mbData = 0;
   if( tbc.contentLength < the_falcon_config->maxMemUpload )
   {
      // we know that tbc.contentLength >> than the upload size.
      mbData = new Falcon::MemBuf_1( tbc.contentLength );
   }
   else
   {
      // try to create a temp file
      strcpy( fname, the_falcon_config->uploadDir ); // we know it can't be larger.
      int fname_len = strlen( fname );
      if( fname[fname_len-1] != '/' && fname[fname_len-1] != '\\' )
      {
         fname[fname_len] = '/';
         fname_len++;
      }
      
      // no need to add a final 0, as we're rewriting on it.
      s_random_filename( fname + fname_len, 16 );

      // try to create the file
      apr_status_t result = apr_file_open( &fp_temp, fname,
         APR_READ | APR_WRITE | APR_CREATE | APR_BUFFERED | APR_BINARY,
         APR_OS_DEFAULT,
         r->pool );
      
      if ( result != 0 )
      {
         ap_log_perror( APLOG_MARK, APLOG_ERR, result, r->pool,
            "Can't create temporary file \"%s\" (%d)",
            fname, result );
         upld->setProperty( "error", new Falcon::CoreString( "Can't create the temporary file" ) );
         return false;
      }

      // save the tempfile
      ApacheVM* avm = static_cast<ApacheVM*>(vm);
      avm->addTempFile( fname );

      // we'll have our storage set.
      Falcon::CoreString *storage = new Falcon::CoreString( fname );
      storage->bufferize();
      upld->setProperty( "storage", storage );
   }

   bool bLast = false;

   // read each line and parse the needed fields. -- first, in speculative mode
   while ( !bLast &&
      (ap_get_brigade(r->input_filters, brigade, AP_MODE_SPECULATIVE, APR_BLOCK_READ, len) == APR_SUCCESS)
      )
   {                  
      apr_brigade_flatten(brigade, buf, &len);
      apr_brigade_cleanup(brigade);

      if ( len == 0 ) // have we exhausted the stream?
      {
         upld->setProperty( "error", new Falcon::CoreString( "Malformed multipart document" ) );
         return false;
      }
      
      // enough input for a full boundary?
      if ( len >= bl )
      {
         buf[len] = 0; // or strstr will loop.
         
         // check if we have a boundary.
         const char* tok = (const char *) memchr( buf, boundary[0], len - bl );
         while( tok != 0 )
         {
            if( memcmp( tok, boundary, bl ) == 0 )
            {
               // trim to just before the boundary.
               len = tok - buf;
               bLast = true;
               break;
            }
            else {
               // try again
               tok++;
               tok = (const char *) memchr( tok, boundary[0], len - ( tok - buf ) - bl );
            }
         }

         if ( ! bLast )
         {
            // ok, we can speculate that there is no boundary in this chain.
            // but the boundary may be hidden right past our last check.
            len -= bl-1;
         }

         // otherwise, the next read will consume everything and we'll be out with an error.
      }

      // Now we know exactly what to read and consume.
      if( ap_get_brigade(r->input_filters, brigade, AP_MODE_READBYTES, APR_BLOCK_READ, len) != APR_SUCCESS )
      {
         upld->setProperty( "error", new Falcon::CoreString( "Error while reading from remote" ) );
         return false;
      }
      apr_brigade_cleanup(brigade);

      // store the data to our media
      if ( mbData != 0 )
      {
         memcpy( mbData->data() + tlen, buf, len );
      }
      else
      {
         apr_size_t wrtlen = len;
         apr_status_t result =  apr_file_write( fp_temp, buf, &wrtlen );
         if ( result != 0 )
         {
            ap_log_perror( APLOG_MARK, APLOG_ERR, result, r->pool,
               "Can't write on temporary file \"%s\" (%d)",
               fname, result );
            upld->setProperty( "error", new Falcon::CoreString( "Can't write on the temporary file" ) );
            apr_file_close( fp_temp );
            return false;
         }
      }

      // and account length
      tlen += len;
      len = 2046;
   }

   // store the data on the final media.
   if( mbData != 0 )
   {
      #if FALCON_VERSION_NUM > 0x00080C
      mbData->resize( (Falcon::uint32) tlen );
      #else
      Falcon::MemBuf *mb = mbData;
      mbData = new Falcon::MemBuf_1( (Falcon::uint32) tlen );
      memcpy( mbData->data(), mb->data(), tlen );
      #endif

      upld->setProperty( "data", mbData );
   }
   else
   {
      apr_file_close( fp_temp );
      // Storage is already set to the final name.
   }

   upld->setProperty( "size", (Falcon::int64) tlen );

   // consume the last boundary, adding 2 extra bytes
   // the two extra bytes are a meaningless 0A 0D sequence for non-last boundaries
   // and '--' for the last boundary.
   len = bl + 2;
   if( ap_get_brigade(r->input_filters, brigade, AP_MODE_READBYTES, APR_BLOCK_READ, len) != APR_SUCCESS )
   {
      upld->setProperty( "error", new Falcon::CoreString( "Error while reading from remote" ) );
      return false;
   }
   apr_brigade_flatten(brigade, buf, &len);
   apr_brigade_cleanup(brigade);
   if( len > 2 && buf[len-1] == '-' )
      return false;

   // there are more fields.
   return true;
}


// Reads a single field in the request
bool s_read_multipart_field( Falcon::VMachine *vm, struct table_callback &tbc,
                     apr_bucket_brigade *brigade, request_rec *r, Falcon::CoreDict* ld  )
{
   apr_size_t len;
   char buf[1025];
   len = 1024;

   const char* boundary = tbc.boundary;
   int bl = tbc.boundaryLen;

   bool bHeader = true;

   // Decoded values
   Falcon::CoreString* field_s = 0;
   Falcon::CoreString* file_name_s = 0;
   Falcon::CoreString* content_type_s = 0;
   Falcon::CoreString* value_s = 0;

   // read each line and parse the needed fields.
   while (ap_get_brigade(r->input_filters, brigade, AP_MODE_GETLINE, APR_BLOCK_READ, len) == APR_SUCCESS)
   {
      apr_brigade_flatten(brigade, buf, &len);
      apr_brigade_cleanup(brigade);
      if ( !len )
      {
         break;
      }

      buf[len] = 0;

      // is this the boundary? -- we're done, break.
      // we got a line of text -- and it should be the boundary
      const char* tok = strstr( buf, boundary );
      if ( tok != 0 )
      {
         // found -- save the field
         if( field_s != 0 )
         {
            // have we found a value?
            if ( value_s == 0 )
            {
               ld->put( field_s, new Falcon::CoreString );
            }
            else
            {
               // The value has always an extra \r\n inserted before the boundary
               Falcon::uint32 len = value_s->length();
               if ( len >= 2 && value_s->getCharAt( len-2 ) == '\r' && value_s->getCharAt( len-1 ) == '\n' )
                  value_s->remove( len-2, 2 );
               
               ld->put( field_s, value_s );
            }
         }
         
         // -- is this the last?
         if ( tok[bl] == 0 || tok[bl] == '-' )
            return false;

         // there are more fields
         return true;
      }

      // are we parsing the value or the header?
      if ( bHeader )
      {
         // break in fields
         char *value;
         char *key = apr_strtok(buf, ":" ,&value);
         if ( key != 0 )
         {
            if ( apr_strnatcasecmp( key, "Content-Disposition" ) == 0 )
            {
               // value should be "form-data" but we skip the check
               char *rest;
               char *field = apr_strtok(value, ";" ,&rest);
               while( field )
               {
                  char *nfield = apr_pstrdup( r->pool, field );
                  char *sfv;
                  char *sfk = apr_strtok (nfield, "=" ,&sfv );

                  if( apr_strnatcasecmp( sfk, "name" ) == 0 )
                  {
                     // we have found the field name 
                     field_s = s_get_header_field( sfv, vm );
                  }
                  else if ( apr_strnatcasecmp( sfk, "filename" ) == 0 )
                  {
                     // we have found the file name
                     file_name_s = s_get_header_field( sfv, vm );
                  }

                  field = apr_strtok(NULL, ";" ,&rest);
               }
            }
            else if ( apr_strnatcasecmp( key, "Content-Type" ) == 0 )
            {
               content_type_s = s_get_header_field( value, vm );
            }
            else if ( apr_strnatcasecmp( key, "\n\r" ) == 0 )
            {
               // we're done with the field header
               bHeader = false;

               // first part of the value?
               if ( field_s != 0 && file_name_s != 0 )
               {
                  // create the file record
                  Falcon::Item *uploaded_cls = vm->findWKI( "Uploaded" );
                  fassert( uploaded_cls != 0 );
                  fassert( uploaded_cls->isClass() );
                  Falcon::CoreObject* upld = uploaded_cls->asClass()->createInstance();
                  upld->setProperty( "filename", file_name_s );
                  if( content_type_s != 0 )
                     upld->setProperty( "mimeType", content_type_s );

                  ld->put( field_s, upld );

                  return s_read_uploaded_file( vm, tbc, brigade, r, upld );
               }
            }
         }
      }
      else
      {
         if ( value_s == 0 )
         {
            value_s = new Falcon::CoreString;
         }
         Falcon::String line_s;
         if ( len >= 2
            && ( ( buf[len-2] == '\r' && buf[-1] == '\n' )
                ||( buf[len-2] == '\n' && buf[-1] == '\r' ) )
            )
         {
            buf[len-2] == '\n';
            buf[len-1] == '\0';
         }
         
         line_s.fromUTF8( buf );
         value_s->append( line_s );
      }
      
      // read another line.
      len = 1024;
   }

   // on error, return false -- and drop what we were doing
   return false;
}


// now parse the post request.
static Falcon::CoreDict* s_read_multipart( Falcon::VMachine *vm, struct table_callback &tbc, apr_bucket_brigade *brigade, request_rec *r  )
{
   const char* boundary = tbc.boundary;
   int bl = tbc.boundaryLen;
   // prepare the dictionary
   Falcon::CoreDict* ld = new Falcon::CoreDict(new Falcon::LinearDict);
   // read all the fields
   while( s_read_multipart_field(  vm, tbc, brigade, r, ld ) );

   return ld;
}


/*#
   @global scriptName
   @brief Physical name of the loaded main script.

   This is the name of the script that has been loaded and served through the
   web server request. It may be different from the file name that has been
   actually requested, as the web server may perform several mappings.

   scriptPath + "/" + scriptName is granted to give the complete file path
   where the script has been loaded.
*/

/*#
   @global scriptPath
   @brief Physical path of the loaded main script.

   This is the directory where the script that has been served through the
   web server request has been loaded from.
   It may be different from the path requested, as the web server may perform
   several path mappings.

   scriptPath + "/" + scriptName is granted to give the complete file path
   where the script has been loaded.

   This information can be useful to the scripts to know the base directory of
   their main code, and use that to perform relative operations on the filesystem.

   @note scriptPath won't be changed through submodules; it stays the same all through
   one execution.
*/


/*#
   @object Reply
   @brief Reflects the reply that this script has sent (or will send).

   This object contains the header that the system is going to send back
   to the remote client as soon as the first output (through the ">" operator
   or through print functions) is performed. If output has already been
   performed, then the header becomes read-only, and the @b Reply.sent field
   becomes true.

   Falcon template files willing to change the output header should escape
   to Falcon immediately, giving the escape sequence as their very first
   character, as unescaped text is immediately sent to the upstream server.

   A suitable default header with "+200 OK", no cache pragmas and text/html in
   utf-8 encoding is provided by default.

   @note Actually, the contents of this object are read by the integration module
      by property name protocol. In other words, it is not strictly necessary to
      change this object's properties to have them reflected in the output reply.
      In other words, this object may be substituted by another instance of any
      class, having the same fields or part of them.

   @note Content-type and Content-encoding
   @prop status Numeric HTTP reply code. In example, 200 (meaning OK).

   @prop reason Descriptive http reply reason. In the "+200 OK" standard reply,
      it's the "OK" part.

   @prop headers HTTP reply fields. Mangling this fields allow scripts to alter
      the type of data that the server will send to the clients. In example,
      it is possible to transform the output the string in an image by
      setting the field "Content-type" to "image/jpeg". Please, refer to
      HTTP RFC 2616 at http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html
      for further details.

   @prop sent True when the header has been delivered, false if the header is
      still unsent.
*/

FALCON_FUNC  Reply_init( Falcon::VMachine *vm )
{
   //called at VM link step
   Falcon::CoreObject *self = vm->self().asObject();
   self->setProperty( "status", (Falcon::int64) 200 );
   self->setProperty( "reason", new Falcon::CoreString( "OK" ) );

   Falcon::CoreDict *fields = new Falcon::CoreDict( new Falcon::LinearDict( 8 ) );
   self->setProperty( "fields", fields );

   // prepare default values
   fields->put( new Falcon::CoreString( "Content-type" ),
      new Falcon::CoreString( "text/html;charset=utf-8" ) );

   fields->put( new Falcon::CoreString( "Pragma" ),
      new Falcon::CoreString( "no-cache" ) );

   fields->put( new Falcon::CoreString( "Cache-Control" ),
      new Falcon::CoreString( "no-cache" ) );
}

/*#
   @method commit Reply
   @brief Sends immediately the header and pending data.

   Usually, the integration module sends the header upstream
   at a reasonable moment, but some scripts may prefer to send
   the header sooner; in example, this is useful if the script
   is just generating a reply that consists in an HTTP header
   without data, or to start a keep-alive HTTP/1.1 conversation.
*/
FALCON_FUNC  Reply_commit( Falcon::VMachine *vm )
{
   ApacheVM* avm = Falcon::dyncast<ApacheVM*>( vm );
   avm->apacheOutput()->flush();  // will do everything that's needed
}

/*#
   @method setCookie Reply
   @brief sets a cookie that will be received next time that a script is called.
   @param name Name of the Cookie or complete cookie specification in a dictionary.
   @optparam value Value for the cookie (eventually truned into a string).
   @optparam expires Expiration date (a TimeStamp or an ISO or RFC2822 formatted string)
   @optparam path Cookie path validity
   @optparam domain Domain for which the cookie is valid.
   @optparam secure True to specify that the cookie can be sent only in https.
   @optparam httpOnly If true, this cookie will be invisible to client side scripts, and will
             be only sent to the server.

   This facility allows to store variables on the remote system (usually
   a web browser), which will send them back each time it connects again.

   The cookies sent through this function will be received in the @b cookies
   member of subquesent call to this or other scripts.

   Parameters to be left blank can be skipped using @b nil; however, it may be useful
   to use the named parameter convencion, that is, passing a single dictionary containing
   the parameter names and their values.

   In example:
   @code
      Reply.setCookie( "cookie1", "value1", nil, nil, ".mydomain.com", false, true )

      // but probably better
      Reply.setCookie( [
         "name"=>"cookie1",
         "value"=>"value1",
         "domain"=>".mydomain.com",
         "httpOnly"=> true
         ])
   @endcode

   Only the @b name parameter is mandatory.
   @note Cookies must be set before output is sent to the upstream server.
*/
FALCON_FUNC  Reply_setCookie( Falcon::VMachine *vm )
{
   Falcon::String sDummy;
   Falcon::String sCookie;

   Falcon::Item *i_name = vm->param(0);
   if( i_name == 0 || ! (i_name->isDict() || i_name->isString() ) )
   {
      goto invalid;
   }

   Falcon::Item *i_value;
   Falcon::Item *i_expire;
   Falcon::Item *i_path;
   Falcon::Item *i_domain;
   Falcon::Item *i_secure;
   Falcon::Item *i_httpOnly;

   if ( i_name->isDict() )
   {
      Falcon::CoreDict *dict = i_name->asDict();
      Falcon::String tmp = "name";
      i_name = dict->find( &tmp );
      if ( i_name == 0 || ! i_name->isString())
         goto invalid;

      tmp = "value";
      i_value = dict->find( &tmp );
      tmp = "expires";
      i_expire = dict->find( &tmp );
      tmp =  "path";
      i_path = dict->find( &tmp );
      tmp = "domain";
      i_domain = dict->find( &tmp );
      tmp = "secure";
      i_secure = dict->find( &tmp );
      tmp = "httpOnly";
      i_httpOnly = dict->find( &tmp );
   }
   else {
      i_value = vm->param(1);
      i_expire = vm->param(2);
      i_path = vm->param(3);
      i_domain = vm->param(4);
      i_secure = vm->param(5);
      i_httpOnly = vm->param(6);
   }

   // name is a string here.
   Falcon::URI::URLEncode( *i_name->asString(), sCookie );
   sCookie.append( "=" );

   // if value is not a string, stringify it
   if( i_value != 0 )
   {
      Falcon::String temp;

      if ( i_value->isString() )
      {
         Falcon::URI::URLEncode( *i_value->asString(), temp );
      }
      else {
         vm->itemToString( sDummy, i_value );
         Falcon::URI::URLEncode( sDummy, temp );
      }
      sCookie.append( temp );
   }

   // Expire part
   if ( i_expire != 0 )
   {
      // a bit of sanitization; if we have an expire, we must ensure it's in ISO or RFC2822 format.
      if( i_expire->isObject() && i_expire->asObject()->derivedFrom( "TimeStamp" ) )
      {
         sDummy="";
         Falcon::TimeStamp *ts = (Falcon::TimeStamp *) i_expire->asObject()->getUserData();
         ts->toRFC2822( sDummy );
         sCookie += "; expires=";
         sCookie += sDummy;
      }
      else if ( i_expire->isString() )
      {
         Falcon::TimeStamp tsDummy;
         if( Falcon::TimeStamp::fromRFC2822( tsDummy, *i_expire->asString() ) )
         {
            sCookie += "; expires=";
            sCookie += *i_expire->asString();
         }
         else
            goto invalid;
      }
      else if ( ! i_expire->isNil() )
         goto invalid;
   }

   // path part
   if ( i_path != 0 )
   {
      if ( i_path->isString() )
      {
         sCookie += "; path=";
         sCookie += *i_path->asString();
      }
      else if ( ! i_path->isNil() )
         goto invalid;
   }

   if ( i_domain != 0 )
   {
      if ( i_domain->isString() )
      {
         sCookie += "; domain=";
         sCookie += *i_domain->asString();
      }
      else if ( ! i_domain->isNil() )
         goto invalid;
   }

   if( i_secure != 0 && i_secure->isTrue() )
   {
      sCookie += "; secure";
   }

   if( i_httpOnly != 0 && i_httpOnly->isTrue() )
   {
      sCookie += "; httponly";
   }

   // great, we have it.
   {
      Falcon::AutoCString asCookie( sCookie );
      ApacheVM *avm = (ApacheVM *) vm;
      apr_table_add( avm->request()->headers_out, "Set-Cookie", asCookie.c_str() );
   }

   return;

invalid:
   throw new Falcon::ParamError(
      Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).extra( "S|D,[X,TimeStamp|S,S,S,B,B]" ) );
}

/*#
   @object Request
   @brief Main web server interface object.

   Object Request contains the informations that are retreived
   from the web server, and allows to exchange data with that.

   In forms and gets,  If the field name in the request ends with
      "[]", then the entry in the gets dictionary is an array containing all the
      values posted under the same field name

   @prop uri Original request URI. This is the unparsed URI
      that has been requested by the remote host and that
      caused the current program to be invoked. It can be
      passed as the parameter of an URI class to have it
      parsed.

   @prop method Original request method. Can be "GET", "POST",
      or HTTP methods.

   @prop headers Original request headers (in a dictionary).
   @prop cookies Dictionary containing the cookies set in the request.
*/

FALCON_FUNC  Request_init( Falcon::VMachine *vm )
{
   ApacheVM *avm = (ApacheVM *) vm;
   request_rec *request = avm->request();

   Falcon::CoreObject *wwwObj = vm->self().asObject();
   wwwObj->setProperty( "method", new Falcon::CoreString( request->method ) );

   apr_uri_t *uri = &request->parsed_uri;

   //================================================
   // Parse headers 
   
   Falcon::CoreDict *headers = new Falcon::CoreDict( new Falcon::LinearDict );
   Falcon::CoreDict *cookies = new Falcon::CoreDict( new Falcon::LinearDict );
   struct table_callback tbc;
   tbc.avm = avm;
   tbc.headers = headers;
   tbc.cookies = cookies;
   tbc.bIsMultiPart = false;
   tbc.contentLength = 0;

   // parse all the headers
   apr_table_do( table_to_dict, &tbc, request->headers_in, NULL );
   //headers->put( new Falcon::CoreString(vm, "remaining"), request->remaining );

   wwwObj->setProperty( "headers", headers );
   wwwObj->setProperty( "cookies", cookies );

   //================================================
   // Parse get -- this is to be done anyhow.
   
   Falcon::CoreDict *opts;

   if( uri->query != 0 ) {
      Falcon::String query( uri->query );
      opts = avm->parseQuery( query );
   }
   else {
      opts = new Falcon::CoreDict( new Falcon::LinearDict );
   }

   wwwObj->setProperty( "gets", opts );

   //================================================
   // Prepare the post parameters
   if ( request->method_number == M_POST )
   {
      // create a brigade to read data.
      apr_bucket_brigade *brigade = apr_brigade_create(
            request->pool,
            request->connection->bucket_alloc);

      // Now, two things can happen.
      // Multipart data require a more complex handling.
      if (  tbc.bIsMultiPart )
      {
         // but refuse to read the rest if content length is too wide
         if ( the_falcon_config->maxUpload > 0 &&
              the_falcon_config->maxUpload < tbc.contentLength )
         {
            Falcon::String reason = "Upload too large ";
            reason.writeNumber( (Falcon::int64) tbc.contentLength );
            reason += " (max ";
            reason.writeNumber( (Falcon::int64) the_falcon_config->maxUpload );
            reason += ")";

            opts = new Falcon::CoreDict( new Falcon::LinearDict );
            opts->put( new Falcon::CoreString( ":error" ), new Falcon::CoreString( reason ) );
         }
         else
            opts = s_read_multipart( avm, tbc, brigade, request );
      }
      // If the header din't say this is a multipart document, parse straight the input data.
      else {
         Falcon::String post_data;
         s_read_post( post_data, brigade, request );
         opts = avm->parseQuery( post_data );
      }
   }
   else
   {
      opts = new Falcon::CoreDict( new Falcon::LinearDict );
   }

   wwwObj->setProperty( "posts", opts );

   // parses the uri
   Falcon::CoreString *guri = new Falcon::CoreString;

   if( uri->scheme != 0 )
   {
      guri->append( uri->scheme );
   }
   else
      guri->append( "http" );

   guri->append( "://" );

   if( uri->user != 0 )
   {
      guri->append( uri->user );
      if ( uri->password != 0 )
      {
         guri->append( ':' );
         guri->append( uri->password );
      }
      guri->append( '@' );
   }

   guri->append( request->hostname );
   guri->append( request->unparsed_uri );

   wwwObj->setProperty( "uri", guri );
}

/*#
   @property gets Request
   @brief Fields received in the GET request method.

   If the current script is invoked by a query containing query fields in the
   URI, this property contains the a dictionary with the paris
   of key/values contained in the query. Fields whose name end
   with "[]" are translated into arrays and their values is stored
   in the order tey are found in the query.

   In example, if the page is loaded through a form containing the following
   fields:
   @code
     <form action="myscript.fal" method="GET">
     <br/>User id: <input type="text" name="id"/>
     <br/>Hobby: <input type="text" name="hobbies[ ]"/>
     <br/>Hobby: <input type="text" name="hobbies[]"/>
     <br/>Hobby: <input type="text" name="hobbies[]"/>
     </form>
   @endcode

   myscript.fal will receive the following fields in gets:
   @code
      > Request.gets["id"]   // will be the user id
      inspect( Request.gets["hobbies"] )  // will be an array
   @endcode

   Get fields can be generated directly through a query. A link
   to a falcon script followed by "?" and an URL encode query
   will be translated into a GET request, and @b Request.gets fields
   will receive the specified values.

   If a web page contains the following code:
   @code
   <a href="myscript.fal?id=my_user_id&hobbies[]=diving&hobbies[]=collections">
   @endcode
   then, myscript.fal will receive the "id" value and the array specified by hobbies
   in the "hobbies" key of the @b Request.gets property.
*/

/*#
   @property posts Request
   @brief Fields received in the POST method.

   If the current script is invoked through a form declared as having
   a post method, it will receive the values of the form fields.
   Fields whose name end
   with "[]" are translated into arrays and their values is stored
   in the order tey are found in the query.

   In example, if the page is loaded through a form containing the following
   fields:
   @code
     <form action="myscript.fal" method="POST">
     <br/>User id: <input type="text" name="id"/>
     <br/>Hobby: <input type="text" name="hobbies[]"/>
     <br/>Hobby: <input type="text" name="hobbies[]"/>
     <br/>Hobby: <input type="text" name="hobbies[]"/>
     </form>
   @endcode

   myscript.fal will receive the following fields in gets:
   @code
      > Request.posts["id"]   // will be the user id
      inspect( Request.posts["hobbies"] )  // will be an array
   @endcode

   A script may receive both @b gets and @b posts fields if the

*/

/*#
   @method getField Request
   @brief Retreives a query field from either @b Request.gets or @b Request.posts.
   @param field The field name to be found.
   @optparam defval Default value to be returned if the field is not found.
   @return A cookie, POST or GET field value (as a string).
   @raise Access error, if no default value is given and field is not found.

   In certain cases, it is useful to retreive a query field no matter if
   it comes from a cookie, the POST part or the GET part of the query. This method
   searches first the @b Request.gets, then @b Request.posts fields and finally
   the @b Request.cookies. If the field is not found, the given @b default value is
   returned; if that parameter is not specified and the field is not found, an
   AccessError is raised.
*/

FALCON_FUNC Request_getfield( Falcon::VMachine *vm )
{
   Falcon::Item *i_key = vm->param( 0 );
   if ( i_key == 0 || ! i_key->isString() )
   {
      throw new Falcon::ParamError(
         Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).extra( "S,[X]" ) );
      return;
   }

   Falcon::CoreObject *self = vm->self().asObject();
   Falcon::Item i_vects;

   // try in gets.
   self->getProperty( "gets", i_vects );
   if( i_vects.isDict() )
   {
      Falcon::Item *val = i_vects.asDict()->find( *i_key );
      if ( val != 0 )
      {
         vm->retval( *val );
         return;
      }
   }

   // no, not in gets. Try in posts.
   self->getProperty( "posts", i_vects );
   if( i_vects.isDict() )
   {
      Falcon::Item *val = i_vects.asDict()->find( *i_key );
      if ( val != 0 )
      {
         vm->retval( *val );
         return;
      }
   }

   // try in cookies
   self->getProperty( "cookies", i_vects );
   if( i_vects.isDict() )
   {
      Falcon::Item *val = i_vects.asDict()->find( *i_key );
      if ( val != 0 )
      {
         vm->retval( *val );
         return;
      }
   }

   // should we raise or return null on error?
   bool bRaise = vm->paramCount() == 1;

   // nothing; should we raise something?
   if  ( bRaise )
   {
      throw new Falcon::AccessError( Falcon::ErrorParam( Falcon::e_arracc, __LINE__ ).extra( "getField" ) );
      return;
   }

   //else, return the default parameter.
   vm->retval( *vm->param(1) );
}

/*#
   @method fwdGet Request
   @brief Forwards the request, creating a suitable query string for a target URI.
   @optparam all If true, add also the POST fields.
   @return An URI encoded string that can be directly used in further get requests.

   This method simplifies the task to create callback to the page that is being
   processed when it's necessary to forward all the fields received to the new request.

   If the @b all parameter is true, also the fields passed as POST fields will be
   forwared through this method.
   
   @note As the get and post fields are not read-only, it is possible to change their
   contents in this object and then call this method to introduce exceptions in forwarding
   the request.
*/

FALCON_FUNC Request_fwdGet( Falcon::VMachine *vm )
{
   Falcon::CoreObject *self = vm->self().asObject();
   Falcon::CoreString *res = new Falcon::CoreString;
   Falcon::Item i_vects;
   
   self->getProperty( "gets", i_vects );
   if( i_vects.isDict() )
   {
      s_dict_as_uri_string( &i_vects.asDict()->items(), *res );
   }

   Falcon::Item *i_all = vm->param( 0 );
   if( i_all != 0 && i_all->isTrue() )
   {
      self->getProperty( "posts", i_vects );
      if( i_vects.isDict() )
      {
         s_dict_as_uri_string( &i_vects.asDict()->items(), *res );
      }
   }

   vm->retval( res );
}

/*#
   @method fwdPost Request
   @brief Forwards the request, creating a set of hidden input fields.
   @optparam all If true, add also the GET fields.
   @return A string containing pre-encoded http hidden input fields.

   This method simplifies the task to create callback to the page that is being
   processed when it's necessary to forward all the fields received to the new request.

   If the @b all parameter is true, also the fields passed as GET fields will be
   forwared through this method.

   @note As the get and post fields are not read-only, it is possible to change their
   contents in this object and then call this method to introduce exceptions in forwarding
   the request.
*/

FALCON_FUNC Request_fwdPost( Falcon::VMachine *vm )
{
   Falcon::CoreObject *self = vm->self().asObject();
   Falcon::StringStream ss;
   Falcon::Item i_vects;

   self->getProperty( "posts", i_vects );
   if( i_vects.isDict() )
   {
      s_dict_as_input_fields( &i_vects.asDict()->items(), &ss );
   }

   Falcon::Item *i_all = vm->param( 0 );
   if( i_all != 0 && i_all->isTrue() )
   {
      self->getProperty( "gets", i_vects );
      if( i_vects.isDict() )
      {
         s_dict_as_input_fields( &i_vects.asDict()->items(), &ss );
      }
   }

   Falcon::CoreString *res = new Falcon::CoreString;
   ss.closeToString( *res );
   vm->retval( res );
}

//==============================================================
// Generic functions
//
/*#
   @funset template_functions Template functions.
   @brief Functions used to configure and transfer templates.
*/

/*#
   @function sendTemplate
   @brief Configures template file and possibly sends it to the remote end.
   @param stream A Falcon stream opened for reading (or a memory string stream).
   @optparam tpd Data for template conversion.
   @optparam inMemory Work in memory and return the result instead sending it.
   @return The configured contents of the file if @b inMemory is true.
   @raise IoError on error reading the file.
   @inset template_functions

   This function reads a text as-is (in binary mode) and flushes its
   contents to the remote side stream.

   If a dictionary is set as template conversion data, the data in the file
   is converted so that strings between a pair of '%' symbols are expanded in
   the text coresponding to the key in the dictionary. In example, if this is
   a template file:
   @code
      My name is %name%, pleased to meet you!
   @endcode

   The @b %name% configurable text may be changed into "John Smith" throught
   the following call:
   @code
      sendTemplate( InputStream("mytemplate.txt"), ["name" => "John Smith"] )
   @endcode

   If a configurable text is not found in the @b tpd dictionary, it is removed.
   The specal sequence '%%' may be used to write a single '%'.

   @note Maximum lenght of template configurable strings is 64.
*/
FALCON_FUNC sendTemplate( Falcon::VMachine *vm )
{
   ApacheVM *avm = (ApacheVM *) vm;

   // Get name of the file.
   Falcon::Item *i_file = vm->param( 0 );
   Falcon::Item *i_dict = vm->param( 1 );
   Falcon::Item *i_inMemory = vm->param( 2 );

   // parameter sanity check.
   if ( i_file == 0 || ! i_file->isObject() || !i_file->asObject()->derivedFrom( "Stream" ) ||
      ( i_dict != 0 && ! ( i_dict->isDict() || i_dict->isNil() ) )
   )
   {
      throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
         extra( "Stream,[D,X]" ) );
   }

   bool bWorkInMem = i_inMemory == 0 ? false : i_inMemory->isTrue();

   Falcon::Stream *outStream;
   if ( bWorkInMem )
   {
      outStream = new Falcon::StringStream;
   }
   else {
      outStream = new ApacheStream( avm->apacheOutput() );
   }

   Falcon::CoreDict *dataDict = i_dict == 0 || i_dict->isNil() ? 0 : i_dict->asDict();

   Falcon::Stream *inStream = (Falcon::Stream *)i_file->asObject()->getUserData();
   Falcon::String text(1024+96); // spare a bit of space for extra templating.
   while ( ! inStream->eof() )
   {
      if ( ! inStream->readString( text, 1024 ) )
      {
         delete outStream;
         throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
               sysError( inStream->lastError() ) );
      }

      // scan for templating
      if( dataDict != 0 )
      {
         Falcon::uint32 pos0=0, pos1, pos2;
         pos1 = text.find( "%" );
         while ( pos1 != Falcon::String::npos )
         {
            // verify the other position is within 64 chars
            pos2 = text.find( "%", pos1+1);

            // did we broke the limit of the read data?
            while( pos2 == Falcon::String::npos && text.length()-pos1 < 64 )
            {
               Falcon::uint32 c;
               if( ! inStream->get( c ) )
               {
                  delete outStream;
                  throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
                        sysError( inStream->lastError() ) );
               }

               text += c;
               if ( c == '%' )
                  pos2 = text.length();
            }

            // ok; now, if we have found it, fine, else just drop it.
            if( pos2 != Falcon::String::npos )
            {
               // write the other part of the text
               outStream->writeString( text, pos0, pos1 );
               if ( pos2 == pos1 + 1 )
                  outStream->writeString( "%" );
               else
               {
                  // find the key.
                  Falcon::String key = text.subString( pos1+1, pos2 );
                  Falcon::Item *i_value = dataDict->find( &key );
                  // write something only if found
                  if ( i_value != 0 )
                  {
                     if ( i_value->isString() )
                     {
                        outStream->writeString( *i_value->asString() );
                     }
                     else {
                        Falcon::String temp;
                        vm->itemToString( temp, i_value );
                        outStream->writeString( temp );
                     }
                  }
               }
               // search next variable
               pos0 = pos2 + 1;
               pos1 = text.find( "%", pos0 );
            }
            else {
               // just write everything that's left.
               outStream->writeString( text, pos0 );
               pos1 = pos2; // will exit loop
            }
         }

         // write the last part
         outStream->writeString( text, pos0 );
      }
      // if not using the dictionary
      else {
         outStream->writeString( text );
      }
   }

   if ( bWorkInMem )
   {
      Falcon::CoreString *gs = new Falcon::CoreString;
      static_cast<Falcon::StringStream *>(outStream)->closeToString(*gs);
      vm->retval( gs );
   }

   delete outStream;
}


/*#
   @function startedAt
   @brief Returns the value of seconds() when the script was started.
   @return A numeric value that can be used to check script performance.

   This function returns a relative time taken when the web server integration system
   gives control to Falcon, before that the script is actually loaded and started.

   This formula:
   @code
      elapsed = seconds() - startedAt()
   @endcode

   gives the time elapsed in processing the script up to that line, including the
   time to setup the VM and start the execution.
*/

FALCON_FUNC startedAt( Falcon::VMachine *vm )
{
   ApacheVM *avm = (ApacheVM *) vm;
   avm->retval( avm->startedAt() );
}


/*#
   @function htmlEscape
   @brief Escapes a string converting HTML special characters.
   @param string The string to be converted.
   @optparam output A stream where to place the output.
   @return A converted string, or nil if writing to a string.

   This function converts special HTML characters to HTML sequences:
   - '<' is converted to '&lt;'
   - '>' is converted to '&gt;'
   - '&' is converted to '&amp;'
   - '"' is converted to '&quot;'

   If an @b output stream is specified, then the output is sent there
   instead being returned in a new string. If the output of this
   function must be sent directly to output, a true value can be
   passed to send data directly to the VM output stream (usually
   linked with web server data stream). This will spare memory and
   CPU.
*/

FALCON_FUNC htmlEscape( Falcon::VMachine *vm )
{
   Falcon::Item *i_str = vm->param(0);
   Falcon::Item *i_output = vm->param(1);

   // return an empty string if nil was given as parameter.
   if( i_str != 0 && i_str->isNil() )
   {
      vm->retval( new Falcon::CoreString );
      return;
   }

   if( i_str == 0 || ! i_str->isString() )
   {
      throw new Falcon::ParamError(
         Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
         extra( "S,[Stream|X]" ) );
   }

   if ( i_output != 0 )
   {
      if( i_output->isObject() && i_output->asObject()->derivedFrom( "Stream" ) )
         internal_htmlEscape_stream( *i_str->asString(),
            (Falcon::Stream *) i_output->asObject()->getUserData() );
      else
         internal_htmlEscape_stream( *i_str->asString(),
            i_output->isTrue() ? vm->stdOut() : vm->stdErr() );
      return;
   }

   Falcon::StringStream encoded;
   internal_htmlEscape_stream( *i_str->asString(), &encoded );

   Falcon::CoreString *gs = new Falcon::CoreString;
   encoded.closeToString( *gs );
   vm->retval( gs );
}

/*#
   @function uriEscape
   @brief Escapes a string converting it into URI convention.
   @param string The string to be converted.
   @return A converted string.

   This function converts the characters so that the input @b string
   can then be appended to URI requests.
   - ' ' is converted to '+'
   - '%' and all the characters above 0x7F are escaped to '%XX' format,
         where XX is an hex number.
*/
FALCON_FUNC apache_ext_uriEscape( Falcon::VMachine *vm )
{
   Falcon::Item *i_str = vm->param(0);

   if( i_str == 0 || ! i_str->isString() )
   {
      throw new Falcon::ParamError(
         Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
         extra( "S" ) );
   }

   Falcon::CoreString *gs = new Falcon::CoreString;
   Falcon::URI::URLEncode( *i_str->asString(), *gs );
   vm->retval( gs );
}

/*#
   @function uriUnescape
   @brief Unescapes a string converting it from URI convention.
   @param string The string to be converted.
   @return A converted string.
   @raise ParseError if the string is not correctly encoded as an URI string.

   @see uriEscape
*/
FALCON_FUNC apache_ext_uriUnescape( Falcon::VMachine *vm )
{
   Falcon::Item *i_str = vm->param(0);

   if( i_str == 0 || ! i_str->isString() )
   {
      throw new Falcon::ParamError(
         Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
         extra( "S" ) );
   }

   Falcon::CoreString *gs = new Falcon::CoreString;
   if ( ! Falcon::URI::URLDecode( *i_str->asString(), *gs ) )
   {
       throw new Falcon::ParseError(
         Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
         extra( "Decoded" ) );
   }

   vm->retval( gs );
}

//============================================
// Uploaded class - class managing uploads.
//============================================

/*#
   @class Uploaded
   @brief Entity storing uploaded file data.

   Forms containing uploadable files are returned in the @a Request.posts
   dictionary. In those forms, the entries coresponding to uploaded
   files are stored as instances of this class.

   The class has support to access the temporary storage where the
   uploaded file has been placed, to read it in memory or to
   access the memory where the system has stored it.

   For more informations read the @a upload_control entry.

   @prop size The size of the uploaded file, in bytes.
   @prop mimeType The declared mime type of the uploaded data. This is
         Determined by the remote browser uploading the file, so it cannot
         be trusted.
   @prop filename The name of the original file before upload. Can be used
         as an hint of the name it should be given locally, or for extra
         confirmation of the mime type.
   @prop storage Complete path to an accessible file in the local filesystem.
         This is where the server has temporarily stored the file. It may
         be destroeyd as soon as the script is closed. If the file was impored
         directly in memory, this field will be nil.
   @prop data Complete content of the file, as a byte-sized MemBuf. If the
         server was setup to store the file in memory as it is received,
         this field will be valorized with a MemBuf containing all the
         data in the file. If this option is not enabled, the property will
         be nil.
*/

/*#
   @method read Uploaded
   @brief Reads an uploaded file from the temporary storage into memory.
   @return True if the file was actually read, false if it was already stored in memory.
   @raise IoError on read error.
   @raise TypeError if the storage property is not a valid filename.
   
   If the uploaded file coresponding to this entry was stored in a temporary
   local file (in the @b storage property), this method reads it
   in a wide-enough MemBuf and stores it in the @b data property.

   It is possible to use this method to make sure that the whole file is
   in the @b data property after a size check.

   @note The server may prevent this operation to be completed if the
   file is too large. Where in doubt, prefer @b Uploaded.open, which has
   the same semantic but that is more flexible and resource-aware.
*/

FALCON_FUNC Uploaded_read( Falcon::VMachine *vm )
{
   Falcon::CoreObject *self = vm->self().asObject();
   Falcon::Item i_data;

   if( self->getProperty( "data", i_data ) &&
      ( i_data.isMemBuf() || i_data.isString() )
   )
   {
      vm->regA().setBoolean( false );
      return;
   }

   Falcon::Item i_storage;
   if( ! self->getProperty( "storage", i_storage ) || ! i_storage.isString() )
   {
      // invalid storage?
      throw new Falcon::TypeError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
         extra( ".storage" ) );
   }

   Falcon::FileStream fs;
   if ( ! fs.open(  *i_storage.asString(),
                     Falcon::BaseFileStream::e_omReadOnly,
                     Falcon::BaseFileStream::e_smShareRead ) )
   {
      throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
         sysError( fs.lastError() ) );
   }

   Falcon::int64 filesize = fs.seekEnd(0);
   Falcon::MemBuf *mb = new Falcon::MemBuf_1( filesize );

   fs.seekBegin( 0 );
   Falcon::int64 readIn = 0;
   while( readIn < filesize )
   {
      Falcon::int32 len = fs.read( mb->data() + readIn, filesize - readIn );
      if ( len < 0 )
      {
         throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
            sysError( fs.lastError() ) );
         // anyhow, try to close to avoid system leaks.
         fs.close();
      }
      readIn += len;
   }

   self->setProperty( "data", mb );
   
   if ( ! fs.close() )
   {
      throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
            sysError( fs.lastError() ) );
   }

   vm->regA().setBoolean( true );
}

/*#
   @method store Uploaded
   @brief Stores the uploaded data into a file.
   @param path The location where to store the file, or an open stream.
   @raise IoError on read or write error.
   @raise TypeError if the storage property is not a valid filename.
   
   If @b data is filled, this method saves
   its contents into the file indicated by the @b path parameter;
   if it was stored to a temporary file, a system file move
   is tried, if it fails, a file copy is tried, and the origin
   file is removed after the copy is succesful. 

   On failure, a relevant IoError is raised, but the operation
   doesn't fail in case the original file cannot be deleted.

   @note This method can be also used to move or copy an arbitrary
   file by storing a path directly in the @b storage
   property.
*/

FALCON_FUNC Uploaded_store( Falcon::VMachine *vm )
{
   Falcon::Item *ip_path = vm->param(0);
   
   if( ip_path == 0
       || ( ! ip_path->isString() && ! ip_path->isOfClass( "Stream" ) )
      )
   {
      throw
            new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
                .extra( "S|Stream" ) );
   }
   
   Falcon::CoreObject *self = vm->self().asObject();

   //===============================================
   // Do we store from data?
   //
   Falcon::Item i_data;
   if( self->getProperty( "data", i_data ) &&
      ( i_data.isMemBuf() || i_data.isString() )
   )
   {
      Falcon::FileStream* tgFile = 0;
      Falcon::Stream *tgStream;

      // try to save to ip_path;
      if ( ip_path->isString() )
      {
         // try to open the file.
         tgFile = new Falcon::FileStream;
         if( ! tgFile->create( *ip_path->asString(),
               Falcon::BaseFileStream::e_aUserRead | Falcon::BaseFileStream::e_aUserWrite ) )
         {
            Falcon::int64 le = tgFile->lastError();
            delete tgFile;
            throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
               sysError( le ) );
         }

         tgStream = tgFile;
      }
      else
         tgStream = static_cast<Falcon::Stream*>( ip_path->asObject()->getUserData() );
      
      Falcon::int64 written = 0;
      Falcon::uint32 len = i_data.isString() ? i_data.asString()->size()
                                                : i_data.asMemBuf()->size();
      Falcon::byte* data = i_data.isString() ? i_data.asString()->getRawStorage()
                                                : i_data.asMemBuf()->data();
      
      
      int wrt = 0;
      while( written < len )
      {
         wrt = tgStream->write( data + written, len - written );
         if( wrt < 0 )
         {
            Falcon::int64 le = tgStream->lastError();
            delete tgFile;
            throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
               sysError( le ) );
         }
         written += wrt;
      }

      if( ! tgStream->close() )
      {
         Falcon::int64 le = tgStream->lastError();
         delete tgFile;  // ok also if 0
         throw new Falcon::IoError(
            Falcon::ErrorParam( Falcon::e_io_error, __LINE__ )
            .sysError( le ) );
      }

      // success
      return;
   }


   // =============================================
   // Do we store from storage?

   Falcon::Item i_storage;
   if( ! self->getProperty( "storage", i_storage ) || ! i_storage.isString() )
   {
      // invalid storage?
      throw new Falcon::TypeError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
            .extra( ".storage" ) );
   }

   const Falcon::String &fname = *i_storage.asString();

   // First, try to rename the file.
   if( ip_path->isString() )
   {
      Falcon::int32 status;
      // fname -> dest
      if ( Falcon::Sys::fal_move( fname, *ip_path->asString(), status ) )
         return; // great, we did it
   }

   Falcon::FileStream infile;
   if( ! infile.open( fname, Falcon::BaseFileStream::e_omReadOnly ) || ! infile.good() )
   {
      throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
               sysError( infile.lastError() ) );
   }
   
   // no way? -- try to copy it.
   Falcon::FileStream* tgFile = 0;
   Falcon::Stream *tgStream;

   // try to save to ip_path;
   if ( ip_path->isString() )
   {
      // try to open the file.
      tgFile = new Falcon::FileStream();
      if( ! tgFile->create( *ip_path->asString(),
               Falcon::BaseFileStream::e_aUserRead | Falcon::BaseFileStream::e_aUserWrite ) )
      {
         Falcon::int64 le = tgStream->lastError();
         delete tgFile;
         throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
            sysError( le ) );
      }

      tgStream = tgFile;
   }
   else
      tgStream = static_cast<Falcon::Stream*>( ip_path->asObject()->getUserData() );

   Falcon::byte buffer[2048];
   int wrt = 0;
   while( ! infile.eof() )
   {
      wrt = infile.read( buffer, 2048 );
      if( wrt < 0 )
      {
         delete tgFile;
         throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
            sysError( infile.lastError() ) );
      }
      
      wrt = tgStream->write( buffer, wrt );
         
      if( wrt < 0 )
      {
         Falcon::int64 le = tgStream->lastError();
         delete tgFile;
         throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
            sysError( le ) );
      }
   }

   // silently try to unlink the source file
   infile.close();
   Falcon::int32 status;
   Falcon::Sys::fal_unlink( fname, status );

   // do not raise again if in error
   if( ! tgStream->close() )
   {
      Falcon::int64 le = tgStream->lastError();
      delete tgFile;
      throw new Falcon::IoError(
         Falcon::ErrorParam( Falcon::e_io_error, __LINE__ )
         .sysError( le ) );
   }

   delete tgFile;
}

/*#
   @method open Uploaded
   @brief Opens a read-only Falcon Stream pointing to the uploaed file.
   @return A Falcon stream.
   @raise IoError on open or write error.
   @raise TypeError if the storage property is not a valid filename.
   
   If @b data is filled, this method creates a memory
   read-only StringStream accessing the data as a file. If it was
   stored in a temporary file named as reported by the @b storage
   property, that file is open in read-only/shared mode.

   This method allows to obtain a valid readable stream no matter if the
   uploaded file was cached in memory or temporarily stored to disk.
*/
FALCON_FUNC Uploaded_open( Falcon::VMachine *vm )
{
   Falcon::CoreObject *self = vm->self().asObject();
   Falcon::Item i_data;

   // try to read the data.
   Falcon::Stream *ret = 0;
   
   if( self->getProperty( "data", i_data ) )
   {
      
      if( i_data.isMemBuf() )
      {
         ret = new Falcon::StringStream();
         Falcon::MemBuf *mb = i_data.asMemBuf();
         ret->write( mb->data(), mb->size() );
         ret->seekBegin(0);
      }
      else if ( i_data.isString() )
      {
         ret = new Falcon::StringStream( *i_data.asString() );
      }
   }

   // try to load from storage if we didn't create a stream.
   if ( ret == 0 )
   {
      Falcon::Item i_storage;
      if( ! self->getProperty( "storage", i_storage ) || ! i_storage.isString() )
      {
         // invalid storage?
         throw new Falcon::TypeError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
               .extra( ".storage" ) );
      }

      Falcon::FileStream *temp = new Falcon::FileStream();
      if ( ! temp->open( *i_storage.asString(), Falcon::BaseFileStream::e_omReadOnly ) )
      {
         Falcon::int64 le = temp->lastError();
         delete temp;
         throw new Falcon::IoError(
            Falcon::ErrorParam( Falcon::e_io_error, __LINE__ )
            .sysError( le ) );
      }
      
      ret = temp;
   }

   // create the stream.
   Falcon::Item *stream_cls = vm->findWKI( "Stream" );
   fassert( stream_cls != 0 );
   fassert( stream_cls->isClass() );
   Falcon::CoreObject *oret = stream_cls->asClass()->createInstance();
   oret->setUserData( ret );
   vm->retval( oret );
}

/*#
   @function tempFile
   @brief Creates a temporary file.
   @optparam name If given and passed by reference, it will receive the complete file name.
   @return A Falcon stream.
   @raise IoError on open or write error.

   Temporary streams are automatically deleted when the the script terminates.

   In case the script want to get the name of the file where the temporary data
   is stored (i.e. to copy or move it elsewhere after having completed the updates),
   the parameter @b name needs to be passed by reference, and it will receive the
   filename.

   @note The temporary files are stored in the directory specified by the
      parameter UploadDir in the falcon.ini file.
*/
FALCON_FUNC apache_ext_tempFile( Falcon::VMachine *vm )
{
   Falcon::Item* i_name = vm->param(0);

   char fname[MAX_UPLOAD_DIR + 20];

   // try to create a temp file
   strcpy( fname, the_falcon_config->uploadDir ); // we know it can't be larger.
   int fname_len = strlen( fname );
   if( fname[fname_len-1] != '/' && fname[fname_len-1] != '\\' )
   {
      fname[fname_len] = '/';
      fname_len++;
   }

   // no need to add a final 0, as we're rewriting on it.
   s_random_filename( fname + fname_len, 16 );

   // try to create the file
   Falcon::FileStream* tgFile = new Falcon::FileStream();
   if( ! tgFile->create( fname,
         Falcon::BaseFileStream::e_aUserRead | Falcon::BaseFileStream::e_aUserWrite ) )
   {
      Falcon::int64 le = tgFile->lastError();
      delete tgFile;
      throw new Falcon::IoError( Falcon::ErrorParam( Falcon::e_io_error, __LINE__ ).
         sysError( le ) );
   }

   // create the stream.
   Falcon::Item *stream_cls = vm->findWKI( "Stream" );
   fassert( stream_cls != 0 );
   fassert( stream_cls->isClass() );
   Falcon::CoreObject *oret = stream_cls->asClass()->createInstance();
   oret->setUserData( tgFile );
   vm->retval( oret );

   // save the tempfile name
   ApacheVM* avm = static_cast<ApacheVM*>(vm);
   avm->addTempFile( fname );

   // eventually report the filename
   if( i_name != 0 && vm->isParamByRef(0) )
   {
      Falcon::CoreString* gs = new Falcon::CoreString( fname );
      gs->bufferize();
      *i_name = gs;
   }
}

//============================================
// Module initialization
//============================================

Falcon::Module * init_apache_ext()
{
   // initialize the module
   Falcon::Module *self = new Falcon::Module();
   self->name( "apache" );
   self->engineVersion( FALCON_VERSION_NUM );
   self->version( VERSION_MAJOR, VERSION_MINOR, VERSION_REVISION );

   /*#
      @global tpd
      @brief Exchange variables with "templatelets".

      This variable can be filled with a dictionary that may be used by
      small Falcon programs (template-lets).
      A typical templatelet:
      @code
         <? names = tpd["names"]
            for name, address, wage in names
            ?><tr><td><?=name></td><tr><td><?=address></td><td><?=wage?></td></tr>
            <? end ?>
      @endcode

      Consider this a common global variable that can be used by topmost programs
      to propagate values downstream.
   */

   self->addGlobal( "tpd", true );

   // Header object
   // create a singleton instance of %Header class.
   Falcon::Symbol *o_header = self->addSingleton( "Reply", Reply_init, true );

   // we don't make it WKS; let it to be exchangeable with another object.
   Falcon::Symbol *c_header = o_header->getInstance();
   self->addClassProperty( c_header, "headers" );
   self->addClassProperty( c_header, "status" );
   self->addClassProperty( c_header, "reason" );
   self->addClassProperty( c_header, "sent" );
   self->addClassMethod( c_header, "commit", Reply_commit );
   self->addClassMethod( c_header, "setCookie", Reply_setCookie ).asSymbol()
      ->addParam( "name" )->addParam( "value" )->addParam( "expires" )
      ->addParam( "path" )->addParam( "domain" )->addParam( "secure" )
      ->addParam( "httpOnly" );

   // create a singleton instance of %Request class (without constructor).
   Falcon::Symbol *o_request = self->addSingleton( "Request", Request_init, true );
   Falcon::Symbol *c_request = o_request->getInstance();
   self->addClassProperty( c_request, "gets" );
   self->addClassProperty( c_request, "posts" );
   self->addClassProperty( c_request, "cookies" );
   self->addClassProperty( c_request, "uri" );
   self->addClassProperty( c_request, "headers" );
   self->addClassProperty( c_request, "method" );
   self->addClassMethod( c_request, "getField", Request_getfield ).asSymbol()
      ->addParam( "field" )->addParam( "defval" );
   self->addClassMethod( c_request, "fwdGet", Request_fwdGet ).asSymbol()
      ->addParam( "all" );
   self->addClassMethod( c_request, "fwdPost", Request_fwdPost ).asSymbol()
      ->addParam( "all" );

   // Create a class for uploaded files
   Falcon::Symbol *c_upfile = self->addClass( "Uploaded" );
   c_upfile->setWKS( true );
   // we don't need an object manager as we don't have internal data.
   self->addClassProperty( c_upfile, "size" );
   self->addClassProperty( c_upfile, "mimeType" );
   self->addClassProperty( c_upfile, "filename" );
   self->addClassProperty( c_upfile, "storage" );
   self->addClassProperty( c_upfile, "data" );
   self->addClassProperty( c_upfile, "error" );
   self->addClassMethod( c_upfile, "read", Uploaded_read );
   self->addClassMethod( c_upfile, "store", Uploaded_store ).asSymbol()
      ->addParam( "path" );
   
   // generic functions
   self->addExtFunc( "sendTemplate", sendTemplate )
      ->addParam( "stream" )->addParam( "tpd" )->addParam( "inMemory" );

   self->addExtFunc( "startedAt", startedAt );

   self->addExtFunc( "htmlEscape", htmlEscape )
      ->addParam( "string" )->addParam( "output" );

   self->addExtFunc( "uriEscape", apache_ext_uriEscape )
      ->addParam( "string" );

   self->addExtFunc( "uriUnscape", apache_ext_uriUnescape )
      ->addParam( "string" );

   self->addExtFunc( "tempFile", apache_ext_tempFile )
      ->addParam( "name" );

   return self;
}

