/*
   FALCON - The Falcon Programming Language.
   FILE: apache_vm.cpp

   Falcon module for Apache 2
   Apache specific Falcon VM.
   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-07-28 18:55:17

   -------------------------------------------------------------------
   (C) Copyright 2008: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

#include "apache_vm.h"

#include <falcon/engine.h>
#include <falcon/dir_sys.h>

#include <httpd.h>
#include <http_protocol.h>
#include <http_log.h>
#include <util_filter.h>
#include "mod_falcon_error.h"

#include <stdio.h>

ApacheVM::TempFileEntry::TempFileEntry( const Falcon::String &fname ):
   m_entry( fname ),
   m_next( 0 )
{
   m_entry.bufferize();
}

ApacheVM::ApacheVM( request_rec *request, ApacheOutput *out, bool binit ):
   VMachine( binit ),
   m_output( out ),
   m_request( request ),
   m_tempFiles( 0 )
{}

ApacheVM::~ApacheVM()
{
}

void ApacheVM::showrefs()
{
   char buffer[60];
   sprintf( buffer, "references %d\n", m_refcount );
   falcon_mod_write_errorstring( m_request, buffer );
}

Falcon::CoreDict *ApacheVM::parseQuery( const Falcon::String &query )
{
   // query is a set of zero or more & separated utf-8 url-encoded strings.

   // we must first find the & chars
   Falcon::String section, key, value;
   Falcon::CoreDict *dict = new Falcon::CoreDict( new Falcon::LinearDict );
   Falcon::uint32 pos1, pos2;

   pos1 = 0;
   pos2 = query.find( "&" );
   while ( pos1 != Falcon::String::npos )
   {
      bool proceed = false;

      // get the substring
      section = query.subString( pos1, pos2 );

      // get the =
      Falcon::uint32 poseq = section.find( "=" );
      if( poseq != Falcon::String::npos )
      {
         if ( Falcon::URI::URLDecode( section.subString( 0, poseq ), key ) &&
              Falcon::URI::URLDecode( section.subString( poseq + 1 ), value )
            )
         {
            proceed = true;
         }

      }
      else
      {
         if( Falcon::URI::URLDecode( section, key ) )
         {
            proceed = true;
            value = "";
         }
      }

      if ( proceed )
      {
          // if the key ends with [], then it's a vector
         if ( key.find( "[]" ) == key.length() -2 )
         {
            key = key.subString( 0, key.length() - 2 );
            // try to find the key in the dictionary...
            Falcon::Item *old_value = dict->find( &key );
            Falcon::CoreArray *array_value;
            if ( old_value == 0 )
            {
               array_value = new Falcon::CoreArray( 5 );
               // add the record (whichy is missing)
               Falcon::CoreString *gkey = new Falcon::CoreString( key );
               gkey->bufferize();
               dict->put( gkey, array_value );
            }
            else if ( ! old_value->isArray() )
            {
               array_value = new Falcon::CoreArray( 5 );
               array_value->append( *old_value );
               // destroy the old value, inserting this one.
               *old_value = array_value;
            }
            else {
               // was already an array
               array_value = old_value->asArray();
            }

            Falcon::CoreString *gvalue = new Falcon::CoreString( value );
            gvalue->bufferize();
            array_value->append( gvalue );
         }
         else {
            // a normal entry
            Falcon::CoreString *gkey = new Falcon::CoreString( key );
            gkey->bufferize();
            Falcon::CoreString *gvalue = new Falcon::CoreString( value );
            gvalue->bufferize();
            dict->put( gkey, gvalue );
         }
      }
      // else, the record was malformed
      // What to do?

      if ( pos2 != Falcon::String::npos )
      {
         pos1 = pos2 + 1;
         pos2 = query.find( "&", pos1 );
      }
      else
         break;
   }

   return dict;
}

void ApacheVM::headerSent()
{
   // tells our header object that data has been sent.
   // supposedly, everything has been correctly setup.
   Falcon::Item *i_header = findGlobalItem( "Reply" );
   // the header may have been changed by the user.
   if ( i_header != 0 && i_header->isObject() )
   {
      Falcon::CoreObject *header = i_header->asObject();
      // and also its content may have been changed.
      header->setProperty( "sent", Falcon::Item(true) );
   }
}

bool ApacheVM::prepareHeader()
{
   // search the item.
   // supposedly, everything has been correctly setup.
   Falcon::Item *i_header = findGlobalItem( "Reply" );

   // the header may have been changed by the user.
   if ( i_header != 0 && i_header->isObject() )
   {
      Falcon::CoreObject *header = i_header->asObject();

      // get the reply code
      Falcon::Item i_status;
      if( header->getProperty( "status", i_status ) && i_status.isOrdinal() )
         m_request->status = i_status.forceInteger();
      // else, leave default

      Falcon::Item i_reason;;
      if( header->getProperty( "reason", i_reason ) && i_reason.isString() )
      {
         // use a buffer, so that status_line can stay valid.
         if( i_reason.asString()->toCString( reply_buffer, 256 ) )
            m_request->status_line = reply_buffer;
      }
      // else, leave default.

      // write the fields.
      Falcon::Item i_fields;
      if( header->getProperty( "headers", i_fields ) && i_fields.isDict() )
      {
         apr_table_t *h = m_request->headers_out;

         Falcon::Iterator di( &i_fields.asDict()->items() );
         while( di.hasCurrent() )
         {
            const Falcon::Item &i_key = di.getCurrentKey();
            const Falcon::Item &i_value = di.getCurrent();
            if( i_key.isString() && i_value.isString() )
            {
               // apr_table_set does not require the strings to stay valid
               Falcon::AutoCString ckey( *i_key.asString() );
               Falcon::AutoCString cval( *i_value.asString() );

               apr_table_set( h, ckey.c_str(), cval.c_str() );

               // also, manage special values
               if( i_key.asString()->compareIgnoreCase( "Content-type" ) == 0 )
               {
                  // use the table entry as we've a persistent storage there.
                  m_request->content_type = apr_table_get( h, ckey.c_str() );
               }
               else if ( i_key.asString()->compareIgnoreCase( "Content-encoding" ) == 0 )
               {
                  m_request->content_encoding = apr_table_get( h, ckey.c_str() );
               }
            }

            di.next();
         }
      }
      else
      {
         // a minimal default
         m_request->content_type = "text/html;charset=utf-8";
      }

      // update the sent field
      header->setProperty( "sent", Falcon::Item(true) );

      return true;
   }
   else {
      // we couldn't setup a proper header, let the system send the defaults
      return false;
   }
}

void ApacheVM::addTempFile( const Falcon::String &fname )
{
   TempFileEntry* tfe = new TempFileEntry( fname );
   tfe->m_next = m_tempFiles;
   m_tempFiles = tfe;
}

void ApacheVM::removeTempFiles( apr_pool_t* p, void* head )
{
      
   TempFileEntry *tfe = (TempFileEntry*) head;
   while( tfe != 0 )
   {
      TempFileEntry *tfe_next = tfe->m_next;
      Falcon::int32 status;
      if ( ! Falcon::Sys::fal_unlink( tfe->m_entry, status ) )
      {
         Falcon::AutoCString cstr( tfe->m_entry );
         ap_log_perror( APLOG_MARK, APLOG_WARNING, status, p,
            "Cannot remove temporary file (%d): %s",
            status, cstr.c_str() );
      }

      delete tfe;
      tfe = tfe_next;
   }
}

/* end of apache_vm.cpp */
