/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_functions.cpp

   Falcon module for Apache 2
   Plugin implementation file
   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-07-28 18:55:17

   -------------------------------------------------------------------
   (C) Copyright 2008: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

#include <stdio.h>
#include <falcon/engine.h>
#include <falcon/dir_sys.h>
#include <falcon/time_sys.h>
#include <falcon/transcoding.h>
#include "mod_falcon.h"
#include "mod_falcon_config.h"
#include "mod_falcon_error.h"
#include "apache_errhand.h"
#include "apache_output.h"
#include "apache_vm.h"
#include "apache_ext.h"
#include "apache_stream.h"
#include "version.h"

#include <http_log.h>
#include <util_filter.h>


//=============================================
// Global pointers to the core and RTL modules
//
static Falcon::Module *s_core =0;
static Falcon::Module *s_ext =0;

static unsigned int s_entropy = 0;

extern "C" {

static int falcon_handler(request_rec *request)
{
   Falcon::String path;

   if ( strcmp(request->handler, FALCON_FAM_TYPE) != 0 &&
        strcmp(request->handler, FALCON_FAL_TYPE) != 0 &&
        strcmp(request->handler, FALCON_FTD_TYPE) != 0)
   {
      return DECLINED;
   }

   // first time in this process?
   if( s_core == 0 )
   {
      Falcon::Engine::Init();
      Falcon::memPool->rampMode( RAMP_MODE_STRICT_ID );

      // create also the core; we know it's empty.
      s_core = Falcon::core_module_init();
      s_ext = init_apache_ext();
   }

   Falcon::numeric startedAt = Falcon::Sys::Time::seconds();

   // verify that the file exists; decline otherwise
   char *script_name = request->filename;
   Falcon::FileStat::e_fileType st;
   if ( ! Falcon::Sys::fal_fileType( script_name, st ) )
   {
      // sorry, the file do not exists, or we cannot access it.
      return DECLINED;
   }

   // we accepted. but are we working?
   if( the_falcon_config->loaded == -1 )
   {
      falcon_mod_write_errorstring( request, "<HTML>\n<BODY>\n<H1>Sorry</H1>\n"
         "<P>Couldn't load Falcon configuration file. See the error log of this server "
         "for further details.</P>\n"
         "</BODY>\n"
         "</HTML>\n"
         );
      return OK;
   }

   // we need some random data
   // for sure, the request address is unique within a time constraint.
   s_entropy += (((long)(startedAt * 1000))%0xFFFFFFFF + ((long) request));
   srand( s_entropy );

   //it's a Falcon module.
   ap_log_rerror( APLOG_MARK, APLOG_INFO, 0, request,
      "Accepting a request from falcon %s", request->filename );

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader;

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   if( the_falcon_config->loadPath[0] != '\0' )
   {
       theLoader.setSearchPath( the_falcon_config->loadPath );
   }
   else
   {
      theLoader.setSearchPath( "." );
      theLoader.addFalconPath();
   }

   // broadcast the same path to the application.
   Falcon::Engine::setSearchPath( theLoader.getSearchPath() );
   
   
   path = theLoader.getSearchPath();
   // for now, let's say we're utf-8
   theLoader.sourceEncoding( "utf-8" );

   // also, instruct the module loader to take the script path as added search path
   Falcon::Path scriptPath( request->canonical_filename );

   if ( scriptPath.isValid() )
   {
      Falcon::String spath;
      scriptPath.getLocation( spath );
      theLoader.addDirectoryFront( spath );
   }

   // Now we need our error handler to manage possible problems in the next steps.
   // First, create the output brigade we'll need to deal with output.
   ApacheOutput aoutput( request );

   // then create an error handler that may use this output structure.
   ApacheErrorHandler errhand( the_falcon_config->errorMode, &aoutput );

   // Allow the script to load iteratively other resources it may need.
   ApacheVM* vm = 0;

   try {

      Falcon::Runtime rt( &theLoader );
      rt.loadFile( script_name );

      // We are ready to go. Let's create our VM and link in minimal stuff
      // we'll handle errors here
      vm = new ApacheVM( request, &aoutput, false ); // request delayed initialization
      vm->startedAt( startedAt );
      vm->stdOut( Falcon::TranscoderFactory( "utf-8", new ApacheStream( &aoutput ), true ));
      vm->stdErr( Falcon::TranscoderFactory( "utf-8", new ApacheStream( &aoutput ), true ));
      // complete initialization
      vm->init();

      // time to tell our apache output who's gonna rule.
      aoutput.setVM( vm );

      // perform link of standard modules.
      //TODO: Lock? -- add a lock provider?
      vm->link( s_core );  // add the core module
      vm->link( s_ext );

      // ok; now prepare the scriptName and scriptPath variables, and set the current working directory
      // so that the script is ready.
      if ( scriptPath.isValid() )
      {
         Falcon::CoreString *spath = new Falcon::CoreString;
         Falcon::CoreString *sfile = new Falcon::CoreString;

         scriptPath.getLocation( *spath );
         scriptPath.getFilename( *sfile );
         Falcon::Item *i_scriptPath = vm->findGlobalItem( "scriptPath" );
         fassert( i_scriptPath != 0 );
         *i_scriptPath = spath;
         Falcon::Item *i_scriptName = vm->findGlobalItem( "scriptName" );
         fassert( i_scriptName != 0 );
         *i_scriptName = sfile;

         int fs;
         Falcon::Sys::fal_chdir( *spath, fs );
      }

      // try to link our module and its dependencies.
      // -- It may fail if there are some undefined symbols
      vm->link( &rt );
      vm->launch();

      // close our stream
      aoutput.close();
   }
   catch( Falcon::Error* error )
   {
      errhand.handleError( error );
      aoutput.close();
   }


   if( vm != 0 )
   {      
      void *temp_files = vm->getTempFiles();
      vm->finalize();
      //For debug: not needed in a real environment
      //Falcon::memPool->performGC();

      if ( temp_files != 0 )
         ApacheVM::removeTempFiles( request->pool, temp_files );
   }

   return OK;
}


//========================================
// Falcon module registration
//
void falcon_register_hook( apr_pool_t *p )
{
   // create falcon standard modules.
   ap_hook_handler(falcon_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

}

/* falcon_functions.cpp */
