/*
   FALCON - The Falcon Programming Language.
   FILE: mod_falcon_config.h

   Falcon specific configuration for the module falcon.
   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: Wed, 30 Apr 2008 18:18:28 +0200

   -------------------------------------------------------------------
   (C) Copyright 2008: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

#include "mod_falcon_config.h"
#include <http_log.h>
#include <apr_strings.h>

extern module AP_MODULE_DECLARE_DATA falcon_module;

falcon_mod_config *the_falcon_config;

//=============================================================
// APACHE hooks
//
void *falcon_mod_create_config(apr_pool_t *p, server_rec *s)
{
   falcon_mod_config *newcfg;
   newcfg = (falcon_mod_config *) apr_pcalloc(p, sizeof(falcon_mod_config));
   newcfg->init_file = DEFAULT_CONFIG_FILE;
   newcfg->pool = p;
   newcfg->loaded = 0;
   newcfg->errorMode = FM_ERROR_MODE_KIND;
   newcfg->maxUpload = FM_DEFAULT_MAX_UPLOAD*1024;
   newcfg->maxMemUpload = 0;
   apr_cpystrn( newcfg->uploadDir, FM_DEFAULT_UPDIR, MAX_UPLOAD_DIR );
   apr_cpystrn( newcfg->loadPath, "", MAX_LOAD_PATH );
   

   // initialize the global instance of our configuration
   the_falcon_config = newcfg;

   return (void *) newcfg;
}

const char *falcon_mod_set_config(cmd_parms *parms, void *mconfig, const char *arg)
{
   falcon_mod_config *cfg = (falcon_mod_config *)
            ap_get_module_config( parms->server->module_config, &falcon_module);

   cfg->init_file = (char *) arg;

   // perform loading
   if ( ! falcon_mod_load_config( the_falcon_config ) )
         cfg->loaded = -1;
      else
         cfg->loaded = 1;
   return NULL;
}

//====================================================================
// Config loader implementation.
//
// We're having a VERY simple config file with just KEY = VALUE pairs.
//
int falcon_mod_load_config( falcon_mod_config *cfg )
{
   char buffer[256];

   ap_log_perror( APLOG_MARK, APLOG_INFO, 0, cfg->pool,
      "Performing lazy initialization of Falcon Config file %s",
      cfg->init_file );

   apr_file_t *fd;
   apr_status_t result = apr_file_open( &fd, cfg->init_file,
         APR_READ | APR_BUFFERED,
         APR_OS_DEFAULT,
         cfg->pool );

   if ( result != 0 )
   {
      ap_log_perror( APLOG_MARK, APLOG_ERR, result, cfg->pool,
         "Can't load Falcon initialization file %s",
         cfg->init_file );
      return 0;
   }

   // Try to load the file.
   result = apr_file_gets( buffer, 255, fd );
   int line = 1;
   while ( ! apr_file_eof( fd ) )
   {
      if ( result != 0 )
      {
         ap_log_perror( APLOG_MARK, APLOG_ERR, result, cfg->pool,
            "Error while reading file %s at line %d",
            cfg->init_file, line );
         apr_file_close( fd );
         return 0;
      }

      apr_collapse_spaces( buffer, buffer );
      // skip lines to be ignored.
      bool correct = true;
      if ( buffer[0] != 0 && buffer[0] != '#' && buffer[0] != ';' )
      {
         // we have a line.
         char *last;
         char *token;
         token = apr_strtok(buffer,"=",&last);
         if( token == NULL )
         {
            correct = false;
         }

         // Log Errors
         else if ( apr_strnatcasecmp( token, "LogErrors" ) == 0 )
         {
            // set error log mode
            token = apr_strtok(NULL,"#",&last);
            if ( token == NULL )
               correct = false;
            else if (apr_strnatcasecmp( token, "Log" ) == 0 )
               cfg->errorMode = FM_ERROR_MODE_LOG;
            else if (apr_strnatcasecmp( token, "Silent" ) == 0 )
               cfg->errorMode = FM_ERROR_MODE_SILENT;
            else if (apr_strnatcasecmp( token, "Kind" ) == 0 )
               cfg->errorMode = FM_ERROR_MODE_KIND;
            else if (apr_strnatcasecmp( token, "Report" ) == 0 )
               cfg->errorMode = FM_ERROR_MODE_REPORT;
            else
               correct = false;
         }
         // Maximum upload size
         else if ( apr_strnatcasecmp( token, "MaxUpload" ) == 0 )
         {
            token = apr_strtok(NULL,"#",&last);
            if ( token == NULL )
               correct = false;
            else {
               int size = atoi( token );
               if ( size == 0 && token[0] != '0' )
                  correct = false;
               else
                  cfg->maxUpload = size*1024;
            }
         }
         // Upload mode
         else if ( apr_strnatcasecmp( token, "MaxMemUpload" ) == 0 )
         {
            token = apr_strtok(NULL,"#",&last);
            if ( token == NULL )
               correct = false;
            else {
               int size = atoi( token );
               if ( size == 0 && token[0] != '0' )
                  correct = false;
               else
                  cfg->maxMemUpload = size*1024;
            }
         }
         // Upload directory
         else if ( apr_strnatcasecmp( token, "UploadDir" ) == 0 )
         {
            token = apr_strtok(NULL,"#",&last);
            if ( token == NULL )
               correct = false;
            else 
               apr_cpystrn( cfg->uploadDir, token, MAX_UPLOAD_DIR );
         }
         // Falcon path
         else if ( apr_strnatcasecmp( token, "LoadPath" ) == 0 )
         {
            token = apr_strtok(NULL,"#",&last);
            if ( token == NULL )
               correct = false;
            else
               apr_cpystrn( cfg->loadPath, token, MAX_LOAD_PATH );
         }
      }

      // Was this line correct?
      if ( ! correct )
      {
         ap_log_perror( APLOG_MARK, APLOG_WARNING, result, cfg->pool,
            "Malformed line in Falcon config file %s at line %d",
            cfg->init_file, line );
      }

      result = apr_file_gets( buffer, 255, fd );
      line++;
   }

   // we're done
   apr_file_close( fd );
   return 1;
}
